@@ -43,22 +43,22 @@ static void     ring_help(void);
 
 static const cmdinfo_t	pop_cmd =
 	{ "pop", NULL, pop_f, 0, 0, 0, NULL,
-	  "pop location from the stack", pop_help };
+	  N_("pop location from the stack"), pop_help };
 static const cmdinfo_t	push_cmd =
-	{ "push", NULL, push_f, 0, 2, 0, "[command]",
-	  "push location to the stack", push_help };
+	{ "push", NULL, push_f, 0, 2, 0, N_("[command]"),
+	  N_("push location to the stack"), push_help };
 static const cmdinfo_t	stack_cmd =
 	{ "stack", NULL, stack_f, 0, 0, 0, NULL,
-	  "view the location stack", stack_help };
+	  N_("view the location stack"), stack_help };
 static const cmdinfo_t  forward_cmd =
 	{ "forward", "f", forward_f, 0, 0, 0, NULL,
-	  "move forward to next entry in the position ring", forward_help };
+	  N_("move forward to next entry in the position ring"), forward_help };
 static const cmdinfo_t  back_cmd =
 	{ "back", "b", back_f, 0, 0, 0, NULL,
-	  "move to the previous location in the position ring", back_help };
+	  N_("move to the previous location in the position ring"), back_help };
 static const cmdinfo_t  ring_cmd =
 	{ "ring", NULL, ring_f, 0, 1, 0, NULL,
-	  "show position ring or move to a specific entry", ring_help };
+	  N_("show position ring or move to a specific entry"), ring_help };
 
 iocur_t	*iocur_base;
 iocur_t	*iocur_top;
@@ -88,7 +88,7 @@ off_cur(
 	int	len)
 {
 	if (iocur_top == NULL || off + len > BBTOB(iocur_top->blen))
-		dbprintf("can't set block offset to %d\n", off);
+		dbprintf(_("can't set block offset to %d\n"), off);
 	else {
 		iocur_top->boff = off;
 		iocur_top->off = ((xfs_off_t)iocur_top->bb << BBSHIFT) + off;
@@ -101,11 +101,17 @@ void
 pop_cur(void)
 {
 	if (iocur_sp < 0) {
-		dbprintf("can't pop anything from I/O stack\n");
+		dbprintf(_("can't pop anything from I/O stack\n"));
 		return;
 	}
-	if (iocur_top->buf)
-		xfree(iocur_top->buf);
+	if (iocur_top->bp) {
+		libxfs_putbuf(iocur_top->bp);
+		iocur_top->bp = NULL;
+	}
+	if (iocur_top->bbmap) {
+		free(iocur_top->bbmap);
+		iocur_top->bbmap = NULL;
+	}
 	if (--iocur_sp >= 0) {
 		iocur_top = iocur_base + iocur_sp;
 		cur_typ = iocur_top->typ;
@@ -128,11 +134,11 @@ pop_f(
 static void
 pop_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " Changes the address and data type to the first entry on the stack.\n"
 "\n"
-		);
+		));
 }
 
 void
@@ -143,18 +149,19 @@ print_iocur(
 	int	i;
 
 	dbprintf("%s\n", tag);
-	dbprintf("\tbyte offset %lld, length %d\n", ioc->off, ioc->len);
-	dbprintf("\tbuffer block %lld (fsbno %lld), %d bb%s\n", ioc->bb,
+	dbprintf(_("\tbyte offset %lld, length %d\n"), ioc->off, ioc->len);
+	dbprintf(_("\tbuffer block %lld (fsbno %lld), %d bb%s\n"), ioc->bb,
 		(xfs_dfsbno_t)XFS_DADDR_TO_FSB(mp, ioc->bb), ioc->blen,
 		ioc->blen == 1 ? "" : "s");
-	if (ioc->use_bbmap) {
-		dbprintf("\tblock map");
-		for (i = 0; i < ioc->blen; i++)
-			dbprintf(" %d:%lld", i, ioc->bbmap.b[i]);
+	if (ioc->bbmap) {
+		dbprintf(_("\tblock map"));
+		for (i = 0; i < ioc->bbmap->nmaps; i++)
+			dbprintf(" %lld:%d", ioc->bbmap->b[i].bm_bn,
+					     ioc->bbmap->b[i].bm_len);
 		dbprintf("\n");
 	}
-	dbprintf("\tinode %lld, dir inode %lld, type %s\n", ioc->ino,
-		ioc->dirino, ioc->typ == NULL ? "none" : ioc->typ->name);
+	dbprintf(_("\tinode %lld, dir inode %lld, type %s\n"), ioc->ino,
+		ioc->dirino, ioc->typ == NULL ? _("none") : ioc->typ->name);
 }
 
 void
@@ -164,11 +171,11 @@ print_ring(void)
 	iocur_t *ioc;
 
 	if (ring_current == -1) {
-		dbprintf("no entries in location ring.\n");
+		dbprintf(_("no entries in location ring.\n"));
 		return;
 	}
 
-	dbprintf("      type    bblock  bblen    fsbno     inode\n");
+	dbprintf(_("      type    bblock  bblen    fsbno     inode\n"));
 
 	i = ring_head;
 	for (;;) {
@@ -222,11 +229,11 @@ push_f(
 		/* check we can execute command */
 		ct = find_command(argv[1]);
 		if (ct == NULL) {
-			dbprintf("no such command %s\n", argv[1]);
+			dbprintf(_("no such command %s\n"), argv[1]);
 			return 0;
 		}
 		if (!ct->canpush) {
-			dbprintf("no push form allowed for %s\n", argv[1]);
+			dbprintf(_("no push form allowed for %s\n"), argv[1]);
 			return 0;
 		}
 	}
@@ -238,7 +245,7 @@ push_f(
 	else
 		set_cur(iocur_top[-1].typ, iocur_top[-1].bb,
 			iocur_top[-1].blen, DB_RING_IGN,
-			iocur_top[-1].use_bbmap ? &iocur_top[-1].bbmap : NULL);
+			iocur_top[-1].bbmap);
 
 	/* run requested command */
 	if (argc>1)
@@ -249,13 +256,13 @@ push_f(
 static void
 push_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " Allows you to push the current address and data type on the stack for\n"
 " later return.  'push' also accepts an additional command to execute after\n"
 " storing the current address (ex: 'push a rootino' from the superblock).\n"
 "\n"
-		);
+		));
 }
 
 /* move forward through the ring */
@@ -266,11 +273,11 @@ forward_f(
 	char		**argv)
 {
 	if (ring_current == -1) {
-		dbprintf("ring is empty\n");
+		dbprintf(_("ring is empty\n"));
 		return 0;
 	}
 	if (ring_current == ring_head) {
-		dbprintf("no further entries\n");
+		dbprintf(_("no further entries\n"));
 		return 0;
 	}
 
@@ -280,8 +287,7 @@ forward_f(
 		iocur_ring[ring_current].bb,
 		iocur_ring[ring_current].blen,
 		DB_RING_IGN,
-		iocur_ring[ring_current].use_bbmap ?
-			&iocur_ring[ring_current].bbmap : NULL);
+		iocur_ring[ring_current].bbmap);
 
 	return 0;
 }
@@ -289,14 +295,14 @@ forward_f(
 static void
 forward_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " The 'forward' ('f') command moves to the next location in the position\n"
 " ring, updating the current position and data type.  If the current location\n"
 " is the top entry in the ring, then the 'forward' command will have\n"
 " no effect.\n"
 "\n"
-		);
+		));
 }
 
 /* move backwards through the ring */
@@ -307,11 +313,11 @@ back_f(
 	char		**argv)
 {
 	if (ring_current == -1) {
-		dbprintf("ring is empty\n");
+		dbprintf(_("ring is empty\n"));
 		return 0;
 	}
 	if (ring_current == ring_tail) {
-		dbprintf("no previous entries\n");
+		dbprintf(_("no previous entries\n"));
 		return 0;
 	}
 
@@ -321,8 +327,7 @@ back_f(
 		iocur_ring[ring_current].bb,
 		iocur_ring[ring_current].blen,
 		DB_RING_IGN,
-		iocur_ring[ring_current].use_bbmap ?
-			&iocur_ring[ring_current].bbmap : NULL);
+		iocur_ring[ring_current].bbmap);
 
 	return 0;
 }
@@ -330,13 +335,13 @@ back_f(
 static void
 back_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " The 'back' ('b') command moves to the previous location in the position\n"
 " ring, updating the current position and data type.  If the current location\n"
 " is the last entry in the ring, then the 'back' command will have no effect.\n"
 "\n"
-		);
+		));
 }
 
 /* show or go to specific point in ring */
@@ -353,8 +358,10 @@ ring_f(
 	}
 
 	index = (int)strtoul(argv[1], NULL, 0);
-	if (index < 0 || index >= RING_ENTRIES)
-		dbprintf("invalid entry: %d\n", index);
+	if (index < 0 || index >= RING_ENTRIES) {
+		dbprintf(_("invalid entry: %d\n"), index);
+		return 0;
+	}
 
 	ring_current = index;
 
@@ -362,7 +369,7 @@ ring_f(
 		iocur_ring[index].bb,
 		iocur_ring[index].blen,
 		DB_RING_IGN,
-		iocur_ring[index].use_bbmap ? &iocur_ring[index].bbmap : NULL);
+		iocur_ring[index].bbmap);
 
 	return 0;
 }
@@ -370,7 +377,7 @@ ring_f(
 static void
 ring_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " The position ring automatically keeps track of each disk location and\n"
 " structure type for each change of position you make during your xfs_db\n"
@@ -389,7 +396,7 @@ ring_help(void)
 " Note: Unlike the 'stack', 'push' and 'pop' commands, the ring tracks your\n"
 " location implicitly.  Use the 'push' and 'pop' commands if you wish to\n"
 " store a specific location explicitly for later return.\n"
-"\n",
+"\n"),
 		RING_ENTRIES);
 }
 
@@ -417,119 +424,55 @@ ring_add(void)
 	}
 }
 
-
-int
-write_bbs(
-	__int64_t       bbno,
-	int             count,
-	void            *bufp,
-	bbmap_t		*bbmap)
+static void
+write_cur_buf(void)
 {
-	int		c;
-	int		i;
-	int		j;
-	int		rval = EINVAL;	/* initialize for zero `count' case */
+	int ret;
 
-	for (j = 0; j < count; j += bbmap ? 1 : count) {
-		if (bbmap)
-			bbno = bbmap->b[j];
-		if (lseek64(x.dfd, bbno << BBSHIFT, SEEK_SET) < 0) {
-			rval = errno;
-			dbprintf("can't seek in filesystem at bb %lld\n", bbno);
-			return rval;
-		}
-		c = BBTOB(bbmap ? 1 : count);
-		i = (int)write(x.dfd, (char *)bufp + BBTOB(j), c);
-		if (i < 0) {
-			rval = errno;
-		} else if (i < c) {
-			rval = -1;
-		} else
-			rval = 0;
-		if (rval)
-			break;
-	}
-	return rval;
+	ret = libxfs_writebufr(iocur_top->bp);
+	if (ret != 0)
+		dbprintf(_("write error: %s\n"), strerror(ret));
+
+	/* re-read buffer from disk */
+	ret = libxfs_readbufr(mp->m_ddev_targp, iocur_top->bb, iocur_top->bp,
+			      iocur_top->blen, 0);
+	if (ret != 0)
+		dbprintf(_("read error: %s\n"), strerror(ret));
 }
 
-int
-read_bbs(
-	__int64_t	bbno,
-	int		count,
-	void		**bufp,
-	bbmap_t		*bbmap)
+static void
+write_cur_bbs(void)
 {
-	void		*buf;
-	int		c;
-	int		i;
-	int		j;
-	int		rval = EINVAL;
+	int ret;
 
-	if (count <= 0)
-		count = 1;
+	ret = libxfs_writebufr(iocur_top->bp);
+	if (ret != 0)
+		dbprintf(_("write error: %s\n"), strerror(ret));
 
-	c = BBTOB(count);
-	if (*bufp == NULL)
-		buf = xmalloc(c);
-	else
-		buf = *bufp;
-	for (j = 0; j < count; j += bbmap ? 1 : count) {
-		if (bbmap)
-			bbno = bbmap->b[j];
-		if (lseek64(x.dfd, bbno << BBSHIFT, SEEK_SET) < 0) {
-			rval = errno;
-			dbprintf("can't seek in filesystem at bb %lld\n", bbno);
-			if (*bufp == NULL)
-				xfree(buf);
-			buf = NULL;
-		} else {
-			c = BBTOB(bbmap ? 1 : count);
-			i = (int)read(x.dfd, (char *)buf + BBTOB(j), c);
-			if (i < 0) {
-				rval = errno;
-				if (*bufp == NULL)
-					xfree(buf);
-				buf = NULL;
-			} else if (i < c) {
-				rval = -1;
-				if (*bufp == NULL)
-					xfree(buf);
-				buf = NULL;
-			} else
-				rval = 0;
-		}
-		if (buf == NULL)
-			break;
-	}
-	if (*bufp == NULL)
-		*bufp = buf;
-	return rval;
+
+	/* re-read buffer from disk */
+	ret = libxfs_readbufr_map(mp->m_ddev_targp, iocur_top->bp, 0);
+	if (ret != 0)
+		dbprintf(_("read error: %s\n"), strerror(ret));
 }
 
 void
 write_cur(void)
 {
-	int ret;
-
 	if (iocur_sp < 0) {
-		dbprintf("nothing to write\n");
+		dbprintf(_("nothing to write\n"));
 		return;
 	}
-	ret = write_bbs(iocur_top->bb, iocur_top->blen, iocur_top->buf,
-		iocur_top->use_bbmap ? &iocur_top->bbmap : NULL);
-	if (ret == -1)
-		dbprintf("incomplete write, block: %lld\n",
-			 (iocur_base + iocur_sp)->bb);
-	else if (ret != 0)
-		dbprintf("write error: %s\n", strerror(ret));
-	/* re-read buffer from disk */
-	ret = read_bbs(iocur_top->bb, iocur_top->blen, &iocur_top->buf,
-		iocur_top->use_bbmap ? &iocur_top->bbmap : NULL);
-	if (ret == -1)
-		dbprintf("incomplete read, block: %lld\n",
-			 (iocur_base + iocur_sp)->bb);
-	else if (ret != 0)
-		dbprintf("read error: %s\n", strerror(ret));
+
+	if (iocur_top->ino_buf)
+		libxfs_dinode_calc_crc(mp, iocur_top->data);
+	if (iocur_top->dquot_buf)
+		xfs_update_cksum(iocur_top->data, sizeof(struct xfs_dqblk),
+				 XFS_DQUOT_CRC_OFF);
+	if (iocur_top->bbmap)
+		write_cur_bbs();
+	else
+		write_cur_buf();
 }
 
 void
@@ -540,26 +483,57 @@ set_cur(
 	int             ring_flag,
 	bbmap_t		*bbmap)
 {
+	struct xfs_buf	*bp;
 	xfs_ino_t	dirino;
 	xfs_ino_t	ino;
 	__uint16_t	mode;
+	const struct xfs_buf_ops *ops = t ? t->bops : NULL;
 
 	if (iocur_sp < 0) {
-		dbprintf("set_cur no stack element to set\n");
+		dbprintf(_("set_cur no stack element to set\n"));
 		return;
 	}
 
-#ifdef DEBUG
-	if (bbmap)
-		printf("xfs_db got a bbmap for %lld\n", (long long)d);
-#endif
+
 	ino = iocur_top->ino;
 	dirino = iocur_top->dirino;
 	mode = iocur_top->mode;
 	pop_cur();
 	push_cur();
-	if (read_bbs(d, c, &iocur_top->buf, bbmap))
+
+	if (bbmap) {
+#ifdef DEBUG_BBMAP
+		int i;
+		printf(_("xfs_db got a bbmap for %lld\n"), (long long)d);
+		printf(_("\tblock map"));
+		for (i = 0; i < bbmap->nmaps; i++)
+			printf(" %lld:%d", (long long)bbmap->b[i].bm_bn,
+					   bbmap->b[i].bm_len);
+		printf("\n");
+#endif
+		iocur_top->bbmap = malloc(sizeof(struct bbmap));
+		if (!iocur_top->bbmap)
+			return;
+		memcpy(iocur_top->bbmap, bbmap, sizeof(struct bbmap));
+		bp = libxfs_readbuf_map(mp->m_ddev_targp, bbmap->b,
+					bbmap->nmaps, 0, ops);
+	} else {
+		bp = libxfs_readbuf(mp->m_ddev_targp, d, c, 0, ops);
+		iocur_top->bbmap = NULL;
+	}
+
+	/*
+	 * Keep the buffer even if the verifier says it is corrupted.
+	 * We're a diagnostic tool, after all.
+	 */
+	if (!bp || (bp->b_error && bp->b_error != EFSCORRUPTED &&
+				   bp->b_error != EFSBADCRC))
 		return;
+	iocur_top->buf = bp->b_addr;
+	iocur_top->bp = bp;
+	if (!ops)
+		bp->b_flags |= LIBXFS_B_UNCHECKED;
+
 	iocur_top->bb = d;
 	iocur_top->blen = c;
 	iocur_top->boff = 0;
@@ -570,18 +544,42 @@ set_cur(
 	iocur_top->ino = ino;
 	iocur_top->dirino = dirino;
 	iocur_top->mode = mode;
-	if ((iocur_top->use_bbmap = (bbmap != NULL)))
-		iocur_top->bbmap = *bbmap;
+	iocur_top->ino_buf = 0;
+	iocur_top->dquot_buf = 0;
 
 	/* store location in ring */
 	if (ring_flag)
 		ring_add();
 }
 
+void
+set_iocur_type(
+	const typ_t	*t)
+{
+	struct xfs_buf	*bp = iocur_top->bp;
+
+	iocur_top->typ = t;
+
+	/* verify the buffer if the type has one. */
+	if (!bp)
+		return;
+	if (!t->bops) {
+		bp->b_ops = NULL;
+		bp->b_flags |= LIBXFS_B_UNCHECKED;
+		return;
+	}
+	if (!(bp->b_flags & LIBXFS_B_UPTODATE))
+		return;
+	bp->b_error = 0;
+	bp->b_ops = t->bops;
+	bp->b_ops->verify_read(bp);
+	bp->b_flags &= ~LIBXFS_B_UNCHECKED;
+}
+
 static void
 stack_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " The stack is used to explicitly store your location and data type\n"
 " for later return.  The 'push' operation stores the current address\n"
@@ -591,7 +589,7 @@ stack_help(void)
 " The 'stack' allows explicit location saves, see 'ring' for implicit\n"
 " position tracking.\n"
 "\n"
-		);
+		));
 }
 
 /*ARGSUSED*/
